port module Main exposing (..)

import Browser
import Browser.Dom as Dom
import Browser.Navigation as Nav
import Cli
import Commands exposing (..)
import Css exposing (..)
import ElmskellTypes.Generated.Decoder exposing (..)
import ElmskellTypes.Generated.Encoder exposing (..)
import ElmskellTypes.Generated.Types exposing (..)
import Html.Styled exposing (Attribute, Html, styled, text, toUnstyled)
import Html.Styled.Attributes exposing (id, value)
import Html.Styled.Events exposing (onInput)
import Json.Decode as D
import Json.Encode as E
import Task
import Types exposing (..)
import Url



-- MAIN


main : Program E.Value Model Msg
main =
    Browser.application
        { init = init
        , view = view
        , update = update
        , subscriptions = subscriptions
        , onUrlChange = UrlChanged
        , onUrlRequest = LinkClicked
        }



-- MODEL


defaultCookies :
    { cookiesKept : CookiesKept
    , font : Font
    , prompt : Prompt
    , theme : Theme
    }
defaultCookies =
    { font = { fontSize = 20.0 }
    , cookiesKept =
        { keepTheme = True
        , keepFont = True
        , keepPrompt = True
        }
    , theme = Dim
    , prompt = { prompt = ">" }
    }


init : E.Value -> Url.Url -> Nav.Key -> ( Model, Cmd Msg )
init flags url key =
    let
        initContent =
            [ text "Welcome to my website! Pardon the alpha quality for the time being"
            , text "\nRun `help` to get started"
            ]

        th =
            case D.decodeValue (D.field "Theme" decodeTheme) flags of
                Ok val ->
                    val

                Err _ ->
                    defaultCookies.theme

        pr =
            case D.decodeValue (D.field "Prompt" decodePrompt) flags of
                Ok val ->
                    val

                Err _ ->
                    defaultCookies.prompt

        cK =
            case D.decodeValue (D.field "CookiesKept" decodeCookiesKept) flags of
                Ok val ->
                    val

                Err _ ->
                    defaultCookies.cookiesKept

        fo =
            case D.decodeValue (D.field "Font" decodeFont) flags of
                Ok val ->
                    val

                Err _ ->
                    defaultCookies.font
    in
    ( { key = key
      , url = url
      , theme = th
      , font = fo
      , cookiesKept = cK
      , prompt = pr
      , content = initContent
      , cliContent = ""
      }
    , Task.attempt (\_ -> NoInitFocus) (Dom.focus "init-focus")
    )



-- UPDATE


update : Msg -> Model -> ( Model, Cmd Msg )
update msg model =
    case msg of
        LinkClicked urlRequest ->
            case urlRequest of
                Browser.Internal url ->
                    ( model, Nav.pushUrl model.key (Url.toString url) )

                Browser.External href ->
                    ( model, Nav.load href )

        UrlChanged url ->
            ( { model | url = url }
            , Cmd.none
            )

        TakeInput string ->
            if String.endsWith "\n" string then
                processInput
                    { model
                        | cliContent = ""
                        , content =
                            model.content
                                ++ [ text
                                        (if model.content /= [] then
                                            "\n"

                                         else
                                            ""
                                        )
                                   , coloredText (coreColor BrightMagenta) model.prompt.prompt
                                   , text <| String.trim string
                                   ]
                    }
                    string

            else
                ( { model | cliContent = string }, Cmd.none )

        NoInitFocus ->
            ( model, Cmd.none )

        ReceivedStorage value ->
            applyJSONData model value


applyJSONData : Model -> E.Value -> ( Model, Cmd Msg )
applyJSONData model data =
    case D.decodeValue (D.field "name" D.string) data of
        Ok "Theme" ->
            case D.decodeValue (D.field "data" decodeTheme) data of
                Ok th ->
                    ( { model | theme = th }, Cmd.none )

                Err e ->
                    errApplyingJSON model e

        Ok "Prompt" ->
            case D.decodeValue (D.field "data" decodePrompt) data of
                Ok pr ->
                    ( { model | prompt = pr }, Cmd.none )

                Err e ->
                    errApplyingJSON model e

        Ok "Font" ->
            case D.decodeValue (D.field "data" decodeFont) data of
                Ok fo ->
                    ( { model | font = fo }, Cmd.none )

                Err e ->
                    errApplyingJSON model e

        Ok _ ->
            ( { model | content = model.content ++ [ text "expecting field `name` to contain type while applyJSONData" ] }, Cmd.none )

        Err e ->
            errApplyingJSON model e


errApplyingJSON : Model -> D.Error -> ( Model, Cmd Msg )
errApplyingJSON model e =
    ( { model | content = model.content ++ [ errMessage <| D.errorToString e ] }, Cmd.none )


isOK : Result x a -> Bool
isOK res =
    case res of
        Ok _ ->
            True

        Err _ ->
            False



-- COMMANDS


processInput : Model -> String -> ( Model, Cmd Msg )
processInput model input =
    let
        cleanInput =
            String.trim input
    in
    case Cli.parseInput commands cleanInput of
        Ok filledCommand ->
            processCommand model filledCommand

        Err _ ->
            ( { model
                | content =
                    model.content
                        ++ [ errMessage "Input not parsed. Come back soon for better error messages."
                           ]
              }
            , Cmd.none
            )


processCommand : Model -> Cli.FilledCommand -> ( Model, Cmd Msg )
processCommand model command =
    if .value <| Cli.getFlag command "--help" then
        helpWith model command

    else
        case command.name of
            "help" ->
                runHelp model command

            "clear" ->
                runClear model command

            "colors" ->
                runColors model command

            "cookies" ->
                runCookies model command

            "debug" ->
                runDebug model command

            "environment" ->
                runEnvironment model command

            "hello" ->
                runHello model command

            "todo" ->
                runTodo model command

            string ->
                ( { model
                    | content = model.content ++ [ warningMessage "This command isn't implemented yet." ]
                  }
                , Cmd.none
                )


runHelp : CommandRunner
runHelp model command =
    let
        snippet_ =
            snippet model
    in
    ( { model
        | content =
            model.content
                ++ (case .value <| Cli.getOption command "command" of
                        Nothing ->
                            [ text <|
                                "\n+---------------------------------------------------------+"
                                    ++ "\n|help       prints this message                           |"
                                    ++ "\n|help COMMAND prints more information about COMMAND       |"
                                    ++ "\n|COMMAND --help is functionally identical to help COMMAND |"
                                    ++ "\n+---------------------------------------------------------+"
                                    ++ "\nclear       clears the screen"
                                    ++ "\ncolors      shows off current color scheme"
                                    ++ "\ncookies     "
                            , broken
                            , text " edits cookie settings"
                            , text "\nenvironment "
                            , unimplemented
                            , text "\nhello       prints Hello, World!"
                            , text "\ntodo        prints my todo list"
                            ]

                        Just "help" ->
                            [ text <|
                                "\n+---------------------------------------------------------+"
                                    ++ "\n|help       prints a list and synopsis of commands        |"
                                    ++ "\n|help COMMAND prints more information about COMMAND       |"
                                    ++ "\n|COMMAND --help is functionally identical to help COMMAND |"
                                    ++ "\n+---------------------------------------------------------+"
                            ]

                        Just "clear" ->
                            [ text <| "\nclear clears the screen"
                            ]

                        Just "colors" ->
                            [ snippet_ "\ncolors show"
                            , text " pretty-prints each color in the current color scheme. It is currently "
                            , unimplemented
                            ]

                        Just "cookies" ->
                            [ newline
                            , newline
                            , broken
                            , text " "
                            , snippet_ "cookies show"
                            , text " works, but no flags actually change settings"
                            , snippet_ "\ncookies show"
                            , text " prints the current cookie policy"
                            , snippet_ "\ncookies [FLAGS]"
                            , text " modifies the persistance of various customization features"
                            , text "\n\nFLAGS:"
                            , text "\n "
                            , snippet_ "--keepTheme=[true|false]"
                            , text " sets whether the theme persists"
                            , text "\n "
                            , snippet_ "--keepPrompt=[true|false]"
                            , text " sets whether the prompt persists"
                            , text "\n "
                            , snippet_ "--keepFont=[true|false]"
                            , text " sets whether the font size persists"
                            , text "\n\nIt's important to note that the name "
                            , snippet_ "cookies"
                            , text "is a fallacy because this site has"
                            , text "\nno cookies. Information about the session is instead stored in the browser's localStorage."
                            , text "\nThe difference is that cookies are sent to the server every time you load the website,"
                            , text "\nwhile localStorage never leaves your computer. I call them cookies to aid normal users,"
                            , text "\nas they are virtually indistinguishable on the user side; when you clear your cookies,"
                            , text "\nyour localStorage gets wiped, too."
                            ]

                        Just "debug" ->
                            [ text "\ndebug is a secret command! Its usage may change any day." ]

                        Just "environment" ->
                            [ text "\nenvironment is ", unimplemented ]

                        Just "hello" ->
                            [ newline
                            , snippet_ "hello"
                            , text " prints Hello, World!"
                            ]

                        Just "todo" ->
                            [ newline
                            , snippet_ "todo"
                            , text " prints my todo list. Just because a feature is on the list doesn't mean it'll be coming any time soon."
                            , text "\nIf you want to see any specific features implemented, contact me via Matrix @mtgmonkey:schildichat.net"
                            , text "\nor via Signal @MTGmonkey. The items on the list are in no particular order and the list is probably"
                            , text "\nout of date."
                            ]

                        Just string ->
                            [ errMessage <| "\nThe command " ++ string ++ " doesn't exist" ]
                   )
      }
    , Cmd.none
    )


runClear : CommandRunner
runClear model command =
    ( { model | content = [] }
    , Cmd.none
    )


runColors : CommandRunner
runColors model command =
    case .value <| Cli.getOption command "show" of
        Nothing ->
            processCommand model { command | flags = [ { name = "--help", value = True } ] }

        Just "show" ->
            ( { model
                | content =
                    model.content
                        ++ [ warningMessage "colors show is ", unimplemented ]
              }
            , Cmd.none
            )

        Just show ->
            ( { model
                | content =
                    model.content
                        ++ [ errMessage <| "colors " ++ show ++ " does not exist. Run "
                           , snippet model "colors --help"
                           ]
              }
            , Cmd.none
            )


runCookies : CommandRunner
runCookies model command =
    let
        flags =
            { keepFont = .value <| Cli.getFlagOption command "--keepFont"
            , keepPrompt = .value <| Cli.getFlagOption command "--keepPrompt"
            , keepTheme = .value <| Cli.getFlagOption command "--keepTheme"
            }
    in
    case .value <| Cli.getOption command "show" of
        Just "show" ->
            ( { model | content = model.content ++ [ text <| "\n" ++ cookiesKeptToString model.cookiesKept ] }, Cmd.none )

        Nothing ->
            let
                keepFont =
                    if flags.keepFont == Just "true" then
                        True

                    else if flags.keepFont == Just "false" then
                        False

                    else
                        model.cookiesKept.keepFont

                keepPrompt =
                    if flags.keepPrompt == Just "true" then
                        True

                    else if flags.keepPrompt == Just "false" then
                        False

                    else
                        model.cookiesKept.keepPrompt

                keepTheme =
                    if flags.keepTheme == Just "true" then
                        True

                    else if flags.keepTheme == Just "false" then
                        False

                    else
                        model.cookiesKept.keepTheme

                cookiesKept =
                    { keepFont = keepFont
                    , keepPrompt = keepPrompt
                    , keepTheme = keepTheme
                    }
            in
            if cookiesKept == model.cookiesKept then
                helpWith model command

            else
                ( { model | cookiesKept = cookiesKept }, saveCookiesKept cookiesKept )

        Just show ->
            ( { model | content = model.content ++ [ errMessage <| "colors " ++ show ++ " does not exist. Run colors --help" ] }, Cmd.none )


runDebug : CommandRunner
runDebug model _ =
    ( { model
        | content =
            model.content
                ++ [ successMessage "You found a secret command!"
                   , text "\ndebug currently does nothing :3"
                   ]
      }
    , Cmd.none
    )


runEnvironment : CommandRunner
runEnvironment model command =
    let
        theme =
            case .value <| Cli.getFlagOption command "--theme" of
                Just th ->
                    case th of
                        "sun" ->
                            Sun

                        "sky" ->
                            Sky

                        "dim" ->
                            Dim

                        "pit" ->
                            Pit

                        _ ->
                            model.theme

                Nothing ->
                    model.theme

        prompt =
            case .value <| Cli.getFlagOption command "--prompt" of
                Just p ->
                    let
                        pr =
                            model.prompt
                    in
                    { pr | prompt = p }

                Nothing ->
                    model.prompt

        font =
            case .value <| Cli.getFlagOption command "--fontsize" of
                Just val ->
                    case String.toFloat val of
                        Just float ->
                            let
                                f =
                                    model.font
                            in
                            { f | fontSize = float }

                        Nothing ->
                            model.font

                Nothing ->
                    model.font
    in
    ( { model
        | prompt = prompt
        , theme = theme
        , font = font
      }
    , Cmd.batch
        [ savePrompt prompt, saveTheme theme, saveFont font ]
    )


runHello : CommandRunner
runHello model _ =
    ( { model | content = model.content ++ [ text "\nHello World!" ] }, Cmd.none )


runTodo : CommandRunner
runTodo model _ =
    ( { model
        | content =
            model.content
                ++
                [ text "\nRun "
                , snippet model "todo --help"
                , text " for more information about this list"
                , text <| "\n- Implement colors throughout existing methods"
                            ++ "\n- Create a strong color language for unified UI"
                            ++ "\n- Finish Colors command. My vision is something like Neofetch"
                            ++ "\n- Collect and store feedback via command line in a database"
                            ++ "\n- Fix cookies so the site respects settings from the cookies command"
                            ++ "\n- Make accessable on mobile"
                            ++ "\n- Add header advert for tor hidden service"
                            ++ "\n- Add plaintext mode for noscript users"
                            ++ "\n- Add warning banner on tor hidden service if Javascript is enabled"
                   ]
      }
    , Cmd.none
    )


helpWith : CommandRunner
helpWith model command =
    processCommand model
        { name = "help"
        , opts = [ { name = "command", value = Just command.name } ]
        , flags = []
        , flagOpts = []
        }



-- COMMAND ABSTRACTIONS


listToString : List String -> String
listToString list =
    "[" ++ String.join "," list ++ "]"


boolToString : Bool -> String
boolToString =
    \n ->
        if n then
            "True"

        else
            "False"


cookiesKeptToString : CookiesKept -> String
cookiesKeptToString cookiesKept =
    "{ keepFont = "
        ++ boolToString cookiesKept.keepFont
        ++ "\n, keepPrompt = "
        ++ boolToString cookiesKept.keepPrompt
        ++ "\n, keepTheme = "
        ++ boolToString cookiesKept.keepTheme
        ++ "\n}"



-- PORTS
-- sets localStorage 'cookies' to E.Value


port setStorage : ( String, E.Value ) -> Cmd a


port getStorage : String -> Cmd a


port receiveStorageFromJS : (E.Value -> msg) -> Sub msg



-- JSON


saveCookiesKept : CookiesKept -> Cmd Msg
saveCookiesKept cookiesKept =
    setStorage ( "CookiesKept", encodeCookiesKept cookiesKept )


saveFont : Font -> Cmd Msg
saveFont font =
    setStorage ( "Font", encodeFont font )


savePrompt : Prompt -> Cmd Msg
savePrompt prompt =
    setStorage ( "Prompt", encodePrompt prompt )


saveTheme : Theme -> Cmd Msg
saveTheme theme =
    setStorage ( "Theme", encodeTheme theme )


loadStorage : Model -> String -> ( Model, Cmd Msg )
loadStorage model key =
    ( model, getStorage key )



-- SUBSCRIPTIONS


subscriptions : Model -> Sub Msg
subscriptions model =
    receiveStorageFromJS ReceivedStorage



-- VIEW


view : Model -> Browser.Document Msg
view model =
    Browser.Document "elmskell"
        [ toUnstyled <| viewBody model ]


viewBody : Model -> Html Msg
viewBody model =
    styledBody model
        []
        [ styledContent model [] model.content
        , styledCL
            model
            []
            [ styledPrompt model [] [ coloredText (coreColor BrightMagenta) model.prompt.prompt ]
            , styledCLI model [ onInput TakeInput, value model.cliContent, id "init-focus" ] []
            ]
        ]



-- STYLES


allColors : Model -> List Color
allColors model =
    List.map
        coreColor
        [ Red
        , Green
        , Yellow
        , Blue
        , Magenta
        , Cyan
        , BrightRed
        , BrightGreen
        , BrightYellow
        , BrightBlue
        , BrightMagenta
        , BrightCyan
        ]
        ++ List.map
            (themeColor model)
            [ Background
            , Foreground
            , Cursor
            , Black
            , White
            , BrightBlack
            , BrightWhite
            ]



-- Colors from Root Loops
-- flavor: intense
-- fruit: raspberry
-- milk: each option
-- sugar: 6
-- colors: 9
-- sogginess: 7


coreColor : CoreColor -> Color
coreColor color =
    case color of
        Red ->
            hex "e14433"

        Green ->
            hex "#359b54"

        Yellow ->
            hex "#a08016"

        Blue ->
            hex "#5a77f2"

        Magenta ->
            hex "#cf2ec8"

        Cyan ->
            hex "1894a5"

        BrightRed ->
            hex "#f36552"

        BrightGreen ->
            hex "#3db361"

        BrightYellow ->
            hex "#b9941a"

        BrightBlue ->
            hex "#7491f8"

        BrightMagenta ->
            hex "#e54ede"

        BrightCyan ->
            hex "#1eabbf"


themeColor : Model -> ThemeColor -> Color
themeColor model color =
    case model.theme of
        Pit ->
            case color of
                Background ->
                    hex "#120211"

                Foreground ->
                    hex "#f3d9f0"

                Cursor ->
                    themeColor model White

                Black ->
                    hex "#380e36"

                White ->
                    hex "#e29edc"

                BrightBlack ->
                    hex "#7c2b77"

                BrightWhite ->
                    hex "#f9ecf7"

        Dim ->
            case color of
                Background ->
                    hex "#380e36"

                Foreground ->
                    hex "#f7e5f4"

                Cursor ->
                    themeColor model White

                Black ->
                    hex "#5e1c56"

                White ->
                    hex "#e8b2e2"

                BrightBlack ->
                    hex "#9d3c98"

                BrightWhite ->
                    hex "#fbf4fa"

        Sky ->
            case color of
                Background ->
                    hex "#f3d9f0"

                Foreground ->
                    hex "#380e36"

                Cursor ->
                    themeColor model White

                Black ->
                    hex "#eec6e9"

                White ->
                    hex "#7c2b77"

                BrightBlack ->
                    hex "#d575cd"

                BrightWhite ->
                    hex "#120211"

        Sun ->
            case color of
                Background ->
                    hex "#f9ecf7"

                Foreground ->
                    hex "#5a1c56"

                Cursor ->
                    themeColor model White

                Black ->
                    hex "#f3d9f0"

                White ->
                    hex "#9d3c98"

                BrightBlack ->
                    hex "#dc8ed5"

                BrightWhite ->
                    hex "#380e36"


styledBody : Model -> List (Attribute Msg) -> List (Html Msg) -> Html Msg
styledBody model =
    styled Html.Styled.main_
        [ backgroundColor <| themeColor model Background
        , color <| themeColor model Foreground
        , minHeight (vh 100)
        , width (vw 100)
        , margin (px 0)
        , padding (px 0)
        ]


styledContent : Model -> List (Attribute Msg) -> List (Html Msg) -> Html Msg
styledContent model =
    styled Html.Styled.span
        [ color <| themeColor model Foreground
        , fontSize (px model.font.fontSize)
        , padding (px 0)
        , margin (px 0)
        , backgroundColor <| themeColor model Background
        , borderWidth (px 0)
        , width (vw 100)
        , whiteSpace preWrap
        , fontFamily monospace
        ]


styledCL : Model -> List (Attribute Msg) -> List (Html Msg) -> Html Msg
styledCL model =
    styled Html.Styled.span
        [ backgroundColor <| themeColor model Background
        , padding (px 0)
        , margin (px 0)
        , displayFlex
        , flexDirection row
        , flexWrap noWrap
        , justifyContent flexStart
        ]


styledCLI : Model -> List (Attribute Msg) -> List (Html Msg) -> Html Msg
styledCLI model =
    styled Html.Styled.textarea
        [ color <| themeColor model Foreground
        , fontSize (px model.font.fontSize)
        , padding (px 0)
        , margin (px 0)
        , backgroundColor <| themeColor model Background
        , borderWidth (px 0)
        , outlineWidth (px 0)
        , height (px model.font.fontSize)
        , resize none
        , overflow hidden
        , flexGrow (Css.int 100)
        ]


styledPrompt : Model -> List (Attribute Msg) -> List (Html Msg) -> Html Msg
styledPrompt model =
    styled Html.Styled.span
        [ color <| themeColor model Foreground
        , fontSize (px model.font.fontSize)
        , padding (px 0)
        , margin (px 0)
        , backgroundColor <| themeColor model Background
        , borderWidth (px 0)
        , height (px model.font.fontSize)
        , fontFamily monospace
        ]


coloredText : Color -> String -> Html Msg
coloredText fgColor string =
    coloredTextWBackground fgColor (rgba 0 0 0 0) string


coloredTextWBackground : Color -> Color -> String -> Html Msg
coloredTextWBackground fgColor bgColor string =
    styled Html.Styled.span
        [ color fgColor
        , backgroundColor bgColor
        , padding (px 0)
        , margin (px 0)
        , borderWidth (px 0)
        ]
        []
        [ text string ]


errMessage : String -> Html Msg
errMessage =
    coloredText (coreColor BrightRed) << (++) "\n[ERR] "


warningMessage : String -> Html Msg
warningMessage =
    coloredText (coreColor Red) << (++) "\n[WARN] "


successMessage : String -> Html Msg
successMessage =
    coloredText (coreColor BrightGreen) << (++) "\n[SUCCESS] "


unimplemented : Html Msg
unimplemented =
    coloredText (coreColor BrightCyan) "[UNIMPLEMENTED]"


broken : Html Msg
broken =
    coloredText (coreColor BrightRed) "[BROKEN]"


snippet : Model -> String -> Html Msg
snippet model =
    coloredTextWBackground (themeColor model BrightWhite) (themeColor model BrightBlack)


newline : Html Msg
newline =
    text "\n"
